///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//  Notice to licensees:                                                     //
//                                                                           //
//  This source code is the exclusive, proprietary intellectual property of  //
//  Sharkysoft (sharkysoft.com).  You may view this source code as a         //
//  supplement to other product documentation, but you may not distribute    //
//  it or use it for any other purpose without written consent from          //
//  Sharkysoft.                                                              //
//                                                                           //
//  You are permitted to modify and recompile this source code, but you may  //
//  not remove this notice.  If you add features to or fix errors in this    //
//  code, please consider sharing your changes with Sharkysoft for possible  //
//  incorporation into future releases of the product.  Thanks!              //
//                                                                           //
//  For more information about Sharkysoft products and services, please      //
//  visit Sharkysoft on the web at                                           //
//                                                                           //
//       http://sharkysoft.com/                                              //
//                                                                           //
//  Thank you for using Lava!                                                //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////



package lava.array;



/******************************************************************************
Casts array elements from one type to another.

<p><b>Details:</b> <code>ArrayCaster</code> casts arrays of primitives from one type to another.  All casting is performed using standard Java primitive narrowing and widening conventions.</p>

<p>Some of the Java primitive widening and narrowing rules are strange and unintuitive.  For these cases, alternate versions are provided in <a href=SignedArrayCaster.html><code>SignedArrayCaster</code></a>.  For unsigned numeric casts, use <a href=UnsignedArrayCaster.html><code>UnsignedArrayCaster</code></a>.</p>

<p>The functions in this class have the following general signature:</p>

<blockquote>
<code><var>DEST_TYPE</var>[] to<var>DEST_TYPE</var> (<var>SRC_TYPE</var>[] <var>src</var>, int <var>from</var>, <var>DEST_TYPE</var>[] <var>dest</var>, int <var>to</var>, int <var>amt</var>)</code>
</blockquote>

<p>where</p>

<dl>
	<dt><var>SRC_TYPE</var>
		<dd>is the primitive type of the source array (byte, short, int, float, etc.)
	<dt><var>src</var>
		<dd>is the source array
	<dt><var>from</var>
		<dd>is the starting index in the source array
	<dt><var>DEST_TYPE</var>
		<dd>is the primitive type of the destination array (byte, short, int, float, etc.)
	<dt><var>dest</var>
		<dd>is the destination array
	<dt><var>to</var>
		<dd>is the starting index in the destination array
	<dt><var>amt</var>
		<dd>is the number of elements from the source array to cast
	<dt>the return value
		<dd>is dest, returned for convenience
</dl>

<p>If <var>src</var> is <code>null</code>, no action is taken and <code>null</code> is returned.  If <var>dest</var> is <code>null</code>, a new array is created whose length is just long enough to accomodate the requested operation, and that new array is returned.</p>

@see SignedArrayCaster
@see UnsignedArrayCaster

@since 1998.05.27
@version 1999.09.11
******************************************************************************/

public abstract class ArrayCaster
{



	public final static double[] toDoubles
	(
		float[] src,
		int from,
		double[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new double [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (double) src [from + i];
		return dest;
	}



	public static double[] toDoubles
	(
		long[] src,
		int from,
		double[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new double [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (double) src [from + i];
		return dest;
	}



	public static double[] toDoubles
	(
		int[] src,
		int from,
		double[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new double [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (double) src [from + i];
		return dest;
	}



	public static double[] toDoubles
	(
		short[] src,
		int from,
		double[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new double [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (double) src [from + i];
		return dest;
	}



	public static double[] toDoubles
	(
		byte[] src,
		int from,
		double[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new double [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (double) src [from + i];
		return dest;
	}



	public final static double[] toDoubles
	(
		char[] src,
		int from,
		double[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new double [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (double) src [from + i];
		return dest;
	}



	public final static float[] toFloats
	(
		double[] src,
		int from,
		float[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new float [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (float) src [from + i];
		return dest;
	}



	public static float[] toFloats
	(
		long[] src,
		int from,
		float[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new float [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (float) src [from + i];
		return dest;
	}



	public static float[] toFloats
	(
		int[] src,
		int from,
		float[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new float [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (float) src [from + i];
		return dest;
	}



	public static float[] toFloats
	(
		short[] src,
		int from,
		float[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new float [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (float) src [from + i];
		return dest;
	}



	public static float[] toFloats
	(
		byte[] src,
		int from,
		float[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new float [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (float) src [from + i];
		return dest;
	}



	public final static float[] toFloats
	(
		char[] src,
		int from,
		float[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new float [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (float) src [from + i];
		return dest;
	}



	public static long[] toLongs
	(
		double[] src,
		int from,
		long[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new long [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (long) src [from + i];
		return dest;
	}



	public static long[] toLongs
	(
		float[] src,
		int from,
		long[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new long [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (long) src [from + i];
		return dest;
	}



	public static long[] toLongs
	(
		int[] src,
		int from,
		long[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new long [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (long) src [from + i];
		return dest;
	}



	public static long[] toLongs
	(
		short[] src,
		int from,
		long[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new long [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (long) src [from + i];
		return dest;
	}



	public static long[] toLongs
	(
		byte[] src,
		int from,
		long[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new long [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (long) src [from + i];
		return dest;
	}



	public final static long[] toLongs
	(
		char[] src,
		int from,
		long[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new long [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (long) src [from + i];
		return dest;
	}



	public static int[] toInts
	(
		double[] src,
		int from,
		int[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new int [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (int) src [from + i];
		return dest;
	}



	public static int[] toInts
	(
		float[] src,
		int from,
		int[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new int [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (int) src [from + i];
		return dest;
	}



	public final static int[] toInts
	(
		long[] src,
		int from,
		int[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new int [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (int) src [from + i];
		return dest;
	}



	public static int[] toInts
	(
		short[] src,
		int from,
		int[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new int [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (int) src [from + i];
		return dest;
	}



	public static int[] toInts
	(
		byte[] src,
		int from,
		int[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new int [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (int) src [from + i];
		return dest;
	}



	public final static int[] toInts
	(
		char[] src,
		int from,
		int[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new int [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (int) src [from + i];
		return dest;
	}



	public static short[] toShorts
	(
		double[] src,
		int from,
		short[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new short [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (short) src [from + i];
		return dest;
	}



	public static short[] toShorts
	(
		float[] src,
		int from,
		short[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new short [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (short) src [from + i];
		return dest;
	}



	public final static short[] toShorts
	(
		long[] src,
		int from,
		short[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new short [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (short) src [from + i];
		return dest;
	}



	public final static short[] toShorts
	(
		int[] src,
		int from,
		short[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new short [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (short) src [from + i];
		return dest;
	}



	public static short[] toShorts
	(
		byte[] src,
		int from,
		short[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new short [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (short) src [from + i];
		return dest;
	}



	public final static short[] toShorts
	(
		char[] src,
		int from,
		short[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new short [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (short) src [from + i];
		return dest;
	}



	public static byte[] toBytes
	(
		double[] src,
		int from,
		byte[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new byte [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (byte) src [from + i];
		return dest;
	}



	public static byte[] toBytes
	(
		float[] src,
		int from,
		byte[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new byte [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (byte) src [from + i];
		return dest;
	}



	public static byte[] toBytes
	(
		long[] src,
		int from,
		byte[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new byte [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (byte) src [from + i];
		return dest;
	}



	public final static byte[] toBytes
	(
		int[] src,
		int from,
		byte[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new byte [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (byte) src [from + i];
		return dest;
	}



	public final static byte[] toBytes
	(
		short[] src,
		int from,
		byte[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new byte [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (byte) src [from + i];
		return dest;
	}



	public final static byte[] toBytes
	(
		char[] src,
		int from,
		byte[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new byte [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (byte) src [from + i];
		return dest;
	}



	public static char[] toChars
	(
		double[] src,
		int from,
		char[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new char [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (char) src [from + i];
		return dest;
	}



	public static char[] toChars
	(
		float[] src,
		int from,
		char[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new char [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (char) src [from + i];
		return dest;
	}



	public final static char[] toChars
	(
		long[] src,
		int from,
		char[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new char [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (char) src [from + i];
		return dest;
	}



	public final static char[] toChars
	(
		int[] src,
		int from,
		char[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new char [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (char) src [from + i];
		return dest;
	}



	public final static char[] toChars
	(
		short[] src,
		int from,
		char[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new char [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (char) src [from + i];
		return dest;
	}



	public static char[] toChars
	(
		byte[] src,
		int from,
		char[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new char [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (char) src [from + i];
		return dest;
	}



	/**********************************************************************
	Identifies primitive array type.

	<p><b>Details:</b>  This method determines which primitive type the elements in this array are, and returns a value as follows:</p>

	<table>
		<tr>
			<th>array type</th>
			<th>returns</th>
		</tr>
		<tr>
			<td>boolean[]</td>
			<td>0</td>
		</tr>
		<tr>
			<td>byte[]</td>
			<td>1</td>
		</tr>
		<tr>
			<td>short[]</td>
			<td>2</td>
		</tr>
		<tr>
			<td>char[]</td>
			<td>3</td>
		</tr>
		<tr>
			<td>int[]</td>
			<td>4</td>
		</tr>
		<tr>
			<td>long[]</td>
			<td>8</td>
		</tr>
		<tr>
			<td>float[]</td>
			<td>6</td>
		</tr>
		<tr>
			<td>double[]</td>
			<td>10</td>
		</tr>
		<tr>
			<td>anything else, including <code>null</code></td>
			<td>-1</td>
		</tr>
	</table>

	<p>Note that for <code>byte[]</code>, <code>short[]</code>, <code>int[]</code>, and <code>long[]</code>, the return value is the size of the element type in bytes.</p>

	@param array the array
	@return the array type if elements are primitive

	@version 1999.08.19
	@since 1999.08.26
	**********************************************************************/

	public static int getPrimitiveElementType (Object array)
	{
		if (array instanceof boolean[])
			return 0;
		if (array instanceof byte[])
			return 1;
		if (array instanceof short[])
			return 2;
		if (array instanceof char[])
			return 3;
		if (array instanceof int[])
			return 4;
		if (array instanceof long[])
			return 8;
		if (array instanceof float[])
			return 6;
		if (array instanceof double[])
			return 10;
		return -1;
	}



}



