///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//  Notice to licensees:                                                     //
//                                                                           //
//  This source code is the exclusive, proprietary intellectual property of  //
//  Sharkysoft (sharkysoft.com).  You may view this source code as a         //
//  supplement to other product documentation, but you may not distribute    //
//  it or use it for any other purpose without written consent from          //
//  Sharkysoft.                                                              //
//                                                                           //
//  You are permitted to modify and recompile this source code, but you may  //
//  not remove this notice.  If you add features to or fix errors in this    //
//  code, please consider sharing your changes with Sharkysoft for possible  //
//  incorporation into future releases of the product.  Thanks!              //
//                                                                           //
//  For more information about Sharkysoft products and services, please      //
//  visit Sharkysoft on the web at                                           //
//                                                                           //
//       http://sharkysoft.com/                                              //
//                                                                           //
//  Thank you for using Lava!                                                //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////



package lava.array;



import lava.konst.LONG;



/******************************************************************************
2<sup>32</sup>-entry virtual byte array.

<p><b>Details:</b>  A <code>ByteArray32</code> is an array of 2<sup>32</sup> <code>byte</code>s.  Every element in the array is available for reading and writing.</p>

<p>Because very few computing environments are equipped to handle physically allocated arrays of this size, <code>ByteArray32</code> uses an on-<wbr>demand page allocation scheme where only the dirty pages are actually allocated.  In other words, segments of the "4 billion byte array" are allocated only as needed.  These segments are allocated in aligned 256-<wbr>byte chunks.  All 4 billion members of the array are "virtually initialized" with a default value at construction time.</p>

<p><code>ByteArray32</code> is ideal for applications that require large, sparsely-<wbr>populated byte arrays where the offsets of the values that are populated tend to cluster in contiguous or nearly contiguous groups.  <code>ByteArray32</code> is efficient because it never requires array resizing, array copying, or address hashing (unlike a growable <code>Vector</code>- or <code>Hashtable</code>-<wbr>based implementation).  <code>ByteArray32</code> also permits a <em>sequential</em> enumeration of occupied memory cells (unlike a <code>Hashtable</code>, which does not produce sequential output).</p>

<p>Array indices for this "virtual byte array" are specified using unsigned <code>int</code>s.  Byte elements can be accessed 1 byte or 4 bytes at a time.  When 4 bytes are accessed simultaneously, the values are combined into a single <code>int</code> structure.  Indices used to access 4 bytes at a time are automatically aligned (ANDed with <code>0xfffffffc</code>).</p>

@since a long time ago
@version 1999.09.13
@author Sharky
******************************************************************************/

public final class ByteArray32
{



	private static final int PAGE_ADDR_SIZE = 24;



	private static final int PAGE_SIZE = 1 << PAGE_ADDR_SIZE;



	private static final int ROUND_MASK = -1 << PAGE_ADDR_SIZE;



	private static final int ADDRESS_MASK = ~ (ROUND_MASK << 8);



	private static int get_page_number (int index)
	{
		return index >>> PAGE_ADDR_SIZE & 0xff;
	}



	private byte default_byte;



	private int default_int;



	private void set_default_int ()
	{
		int b = default_byte & 0xff;
		default_int = b << 24 | b << 16 | b << 8 | b;
	}



	private ByteArray24[] data;



	/**********************************************************************
	Initializes with default value.

	<p><b>Details:</b>  This constructor initializes a new ByteArray32 with all bytes set to the given initial value (<var>default_byte</var>).</p>

	@param default_byte the initial value
	**********************************************************************/

	public ByteArray32 (byte default_byte)
	{
		this . default_byte = default_byte;
		set_default_int ();
	}



	/**********************************************************************
	Retrieves byte.

	<p><b>Details:</b>  <code>get</code> returns the byte at the given address (<var>index</var>).</p>

	@param index (unsigned) the address
	@return the byte
	**********************************************************************/

	public byte get (int index)
	{
		if (data == null)
			return default_byte;
		int page = get_page_number (index);
		if (data [page] == null)
			return default_byte;
		else
			return data [page] . get (index);
	}



	/**********************************************************************
	Sets byte.

	<p><b>Details:</b>  <code>set</code> sets the byte at the given address (<var>index</var>) to the given value (<var>value</var>).</p>

	@param index (unsigned) the address
	@param value the value
	**********************************************************************/

	public void set (int index, byte value)
	{
		if (data == null)
			if (value == default_byte)
				return;
			else
				data = new ByteArray24 [0x100];
		int page = get_page_number (index);
		if (data [page] == null)
			if (value == default_byte)
				return;
			else
				data [page] = new ByteArray24 (default_byte);
		data [page] . set (index, value);
	}



	/**********************************************************************
	Retrieves bytes.

	<p><b>Details:</b>  <code>get</code> retrieves the given range of bytes (<var>index</var>, <var>amount</var>) and writes them into the destination array (<var>dest</var>).</p>

	@param index (unsigned) the address
	@param dest the destination array
	@param dest_from index of first byte in <var>dest</var>
	@param amount number of bytes to retrieve
	**********************************************************************/

	public void get
	(
		int index,
		byte[] dest,
		int dest_from,
		int amount
	)
	{
		if (data == null)
		{
			for (int i = 0; i < amount; ++ i)
				dest [dest_from + i] = default_byte;
			return;
		}
		int page = get_page_number (index);
		while (true)
		{
			final int next_index = (index & ROUND_MASK) + PAGE_SIZE;
			final int new_amount = Math.min (amount, next_index - index);
			if (data [page] == null)
				for (int i = 0; i < new_amount; ++ i)
					dest [dest_from + i] = default_byte;
			else
				data [page] . get (index, dest, dest_from, new_amount);
			amount -= new_amount;
			if (amount == 0)
				return;
			index = next_index;
			dest_from += new_amount;
			++ page;
		}
	}



	/**********************************************************************
	Sets bytes.

	<p><b>Details:</b>  <code>set</code> writes the given sequence of bytes (<var>src</var>, <var>src_from</var>, <var>amount</var>) into this array, starting at the given address (<var>index</var>).</p>

	@param index the address
	@param src the byte sequence
	@param src_from index of first byte in <var>src</var>
	@param amount number of bytes to write
	**********************************************************************/

	public void set
	(
		int index,
		byte[] src,
		int src_from,
		int amount
	)
	{
		if (data == null)
			data = new ByteArray24 [0x100];
		int page = get_page_number (index);
		while (true)
		{
			final int next_index = (index & ROUND_MASK) + PAGE_SIZE;
			final int new_amount = Math.min (amount, next_index - index);
			if (data [page] == null)
				data [page] = new ByteArray24 (default_byte);
			data [page] . set (index, src, src_from, new_amount);
			amount -= new_amount;
			if (amount == 0)
				break;
			index = next_index;
			src_from += new_amount;
			++ page;
		}
	}



	/**********************************************************************
	Retrieves 32-bit word.

	<p><b>Details:</b>  <code>getIntBe</code> returns the 32-<wbr>bit word stored at the given address (<var>index</var>).  The word is assembled using "big endian" byte ordering.</p>

	@param index the address
	@return the word
	**********************************************************************/

	public int getIntBe (int index)
	{
		int page = get_page_number (index);
		if (data == null || data [page] == null)
			return default_int;
		return data [page] . getIntBe (index);
	}



	/**********************************************************************
	Retrieves 32-bit word.

	<p><b>Details:</b>  <code>getIntBe</code> returns the 32-<wbr>bit word stored at the given address (<var>index</var>).  The word is assembled using "little endian" byte ordering.</p>

	@param index the address
	@return the word
	**********************************************************************/

	public int getIntLe (int index)
	{
		int page = get_page_number (index);
		if (data == null || data [page] == null)
			return default_int;
		return data [page] . getIntLe (index);
	}



	/**********************************************************************
	Sets 32-bit word.

	<p><b>Details:</b>  <code>setIntBe</code> sets the 32-<wbr>bit word at the given address (<var>index</var>).  The word is broken into bytes and the bytes are written in "big endian" byte order.</p>

	@param index the address
	@param value the word
	**********************************************************************/

	public void setIntBe (int index, int value)
	{
		if (data == null)
			if (value == default_int)
				return;
			else
				data = new ByteArray24 [0x100];
		int page = get_page_number (index);
		if (data [page] == null)
			if (value == default_int)
				return;
			else
				data [page] = new ByteArray24 (default_byte);
		data [page] . setIntBe (index, value);
	}



	/**********************************************************************
	Sets 32-bit word.

	<p><b>Details:</b>  <code>setIntLe</code> sets the 32-<wbr>bit word at the given address (<var>index</var>).  The word is broken into bytes and the bytes are written in "little endian" byte order.</p>

	@param index the address
	@param value the word
	**********************************************************************/

	public void setIntLe (int index, int value)
	{
		if (data == null)
			if (value == default_int)
				return;
			else
				data = new ByteArray24 [0x100];
		int page = get_page_number (index);
		if (data [page] == null)
			if (value == default_int)
				return;
			else
				data [page] = new ByteArray24 (default_byte);
		data [page] . setIntLe (index, value);
	}



	/**********************************************************************
	Fills array.

	<p><b>Details:</b>  <code>fill</code> writes the given byte value (<var>value</var>) into the given span of bytes (<var>index</var>, <var>amount</var>).  For fill amounts greater than 256, this method can potentially reduce the storage requirements for this virtual array, especially if the fill value is the same as the default value by which this array was constructed.</p>

	@param index the address
	@param amount the number of bytes to fill
	@param value the value to write
	**********************************************************************/

	public void fill (int index, int amount, byte value)
	{
		if (data == null)
			if (value == default_byte)
				return;
			else
				data = new ByteArray24 [0x100];
		int page = get_page_number (index);
		while (true)
		{
			final int next_index = (index & ROUND_MASK) + PAGE_SIZE;
			final int new_amount = Math.min (amount, next_index - index);
			if (new_amount == PAGE_SIZE)
				if (value == default_byte)
					data [page] = null;
				else
					data [page] = new ByteArray24 (value);
			else
			{
				if (data [page] == null)
					data [page] = new ByteArray24 (default_byte);
				data [page] . fill (index, new_amount, value);
			}
			amount -= new_amount;
			if (amount == 0)
				break;
			index = next_index;
			++ page;
		}
	}



	/**********************************************************************
	Searches for value.

	<p><b>Details:</b>  <code>search</code> scans the array for the given value (<var>value</var>), beginning at the given address (<var>index</var>) and searching through <b>amount</b> bytes.  If the value is found, its address is returned.  Otherwise, -1 is returned.</p>

	@param index the address of the first byte
	@param amount number of bytes to search
	@param value the value to seach for
	@return the address of the first matching byte
	**********************************************************************/

	public long search (int index, int amount, byte value)
	{
		if (amount <= 0)
			return -1;
		if (data == null)
			if (value == default_byte)
				return index & ADDRESS_MASK;
			else
				return -1;
		int page = get_page_number (index);
		while (true)
		{
			final int next_index = (index & ROUND_MASK) + PAGE_SIZE;
			final int new_amount = (int) Math.min (amount & LONG.MAX_UINT, next_index - index);
			if (data [page] == null)
			{
				if (value == default_byte)
					return index & LONG.MAX_UINT;
			}
			else
			{
				int result = data [page] . search (index, new_amount, value);
				if (result != -1)
					return page << PAGE_ADDR_SIZE | result;
			}
			amount -= new_amount;
			if (amount == 0)
				break;
			index = next_index;
			++ page;
		}
		return -1;
	}



	/**********************************************************************
	Reduces storage requirements.

	<p><b>Details:</b>  <code>compact</code> scans this array and, if possible, reduces its physical memory requirements by unallocating physical storage for pages whose byte values are all the same.  If all of the bytes in this array are the same value, that value is returned.  Otherwise, -1 is returned.</p>

	@return value of all bytes
	**********************************************************************/

	public int compact ()
	{
		if (data == null)
			return default_byte;
		int b;
		if (data [0] == null)
			b = default_byte;
		else
			b = data [0] . compact ();
		for (int i = 1; i < 0x100; ++ i)
		{
			int c;
			if (data [i] == null)
				c = default_byte;
			else
			{
				c = data [i] . compact ();
				if (c == (default_byte & 0xff))
					data [i] = null;
			}
			if (b != c)
				b = -1;
		}
		if (b >= 0)
		{
			default_byte = (byte) b;
			set_default_int ();
			data = null;
		}
		return b;
	}



}



