///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//  Notice to licensees:                                                     //
//                                                                           //
//  This source code is the exclusive, proprietary intellectual property of  //
//  Sharkysoft (sharkysoft.com).  You may view this source code as a         //
//  supplement to other product documentation, but you may not distribute    //
//  it or use it for any other purpose without written consent from          //
//  Sharkysoft.                                                              //
//                                                                           //
//  You are permitted to modify and recompile this source code, but you may  //
//  not remove this notice.  If you add features to or fix errors in this    //
//  code, please consider sharing your changes with Sharkysoft for possible  //
//  incorporation into future releases of the product.  Thanks!              //
//                                                                           //
//  For more information about Sharkysoft products and services, please      //
//  visit Sharkysoft on the web at                                           //
//                                                                           //
//       http://sharkysoft.com/                                              //
//                                                                           //
//  Thank you for using Lava!                                                //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////



package lava.array;



import java.math.BigDecimal;
import lava.konst.BYTE;
import lava.konst.CHAR;
import lava.konst.DOUBLE;
import lava.konst.FLOAT;
import lava.konst.INT;
import lava.konst.LONG;
import lava.konst.SHORT;



/******************************************************************************
Casts arrays of unsigned primitives.

<p><b>Details:</b>  <code>UnsignedArrayCaster</code> is a collection of functions that cast arrays of <em>unsigned</em> primitives from one type to another using standard <em>Lava</em> primitive narrowing and widening conventions.  For casting operations involving <em>signed</em> types, see <a href=ArrayCaster.html><code>ArrayCaster</code></a> and <a href=SignedArrayCaster.html><code>SignedArrayCaster</code></a>.</p>

<p>The signatures for functions in this class follow the same rules as described in {@link ArrayCaster <code>ArrayCaster</code>}.</p>

@see ArrayCaster
@see SignedArrayCaster

@author Sharky
@version 1999.09.11
******************************************************************************/

public abstract class UnsignedArrayCaster
	extends ArrayCaster
{



	public final static double[] toDoubles
	(
		long[] src,
		int from,
		double[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new double [to + amt];
		for (int i = 0; i < amt; ++ i)
		{
			long s = src [from + i];
			if ((s & LONG.MIN_SLONG) == 0)
				dest [to + i] = (double) s;
			else
				dest [to + i] = (double) (s >>> 1) * 2 + (s & 1);
		}
		return dest;
	}



	public final static double[] toDoubles
	(
		int[] src,
		int from,
		double[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new double [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (double) (src [from + i] & LONG.MAX_UINT);
		return dest;
	}



	public final static double[] toDoubles
	(
		short[] src,
		int from,
		double[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new double [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (double) (src [from + i] & INT.MAX_USHORT);
		return dest;
	}



	public final static double[] toDoubles
	(
		byte[] src,
		int from,
		double[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new double [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (double) (src [from + i] & INT.MAX_UBYTE);
		return dest;
	}



	public final static float[] toFloats
	(
		long[] src,
		int from,
		float[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new float [to + amt];
		for (int i = 0; i < amt; ++ i)
		{
			long s = src [from + i];
			if ((s & LONG.MIN_SLONG) == 0)
				dest [to + i] = (float) s;
			else
				dest [to + i] = (float) ((double) (s >>> 1) * 2 + (s & 1));
		}
		return dest;
	}



	public final static float[] toFloats
	(
		int[] src,
		int from,
		float[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new float [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (float) (src [from + i] & LONG.MAX_UINT);
		return dest;
	}



	public final static float[] toFloats
	(
		short[] src,
		int from,
		float[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new float [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (float) (src [from + i] & INT.MAX_USHORT);
		return dest;
	}



	public final static float[] toFloats
	(
		byte[] src,
		int from,
		float[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new float [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (float) (src [from + i] & INT.MAX_UBYTE);
		return dest;
	}



	public final static long[] toLongs
	(
		double[] src,
		int from,
		long[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new long [to + amt];
		for (int i = 0; i < amt; ++ i)
		{
			double s = src [from + i];
			if (s >= DOUBLE.MAX_ULONG)
				dest [to + i] = LONG.MAX_ULONG;
			else if (s >= DOUBLE.MAX_SLONG)
				dest [to + i] = new BigDecimal (s) . toBigInteger () . longValue ();
			else if (s > 0)
				dest [to + i] = (long) s;
			else
				dest [to + i] = 0;
		}
		return dest;
	}



	public final static long[] toLongs
	(
		float[] src,
		int from,
		long[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new long [to + amt];
		for (int i = 0; i < amt; ++ i)
		{
			float s = src [from + i];
			if (s >= FLOAT.MAX_ULONG)
				dest [to + i] = LONG.MAX_ULONG;
			else if (s >= FLOAT.MAX_SLONG)
				dest [to + i] = new BigDecimal (s) . toBigInteger () . longValue ();
			else if (s > 0)
				dest [to + i] = (long) s;
			else
				dest [to + i] = 0;
		}
		return dest;
	}



	public final static long[] toLongs
	(
		int[] src,
		int from,
		long[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new long [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = src [from + i] & LONG.MAX_UINT;
		return dest;
	}



	public final static long[] toLongs
	(
		short[] src,
		int from,
		long[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new long [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = src [from + i] & LONG.MAX_USHORT;
		return dest;
	}



	public final static long[] toLongs
	(
		byte[] src,
		int from,
		long[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new long [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = src [from + i] & LONG.MAX_UBYTE;
		return dest;
	}



	public final static int[] toInts
	(
		double[] src,
		int from,
		int[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new int [to + amt];
		for (int i = 0; i < amt; ++ i)
		{
			double s = src [from + i];
			if (s >= DOUBLE.MAX_UINT)
				dest [to + i] = INT.MAX_UINT;
			else if (s >= DOUBLE.MAX_SINT)
				dest [to + i] = (int) (long) s;
			else if (s > 0)
				dest [to + i] = (int) s;
			else
				dest [to + i] = 0;
		}
		return dest;
	}



	public final static int[] toInts
	(
		float[] src,
		int from,
		int[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new int [to + amt];
		for (int i = 0; i < amt; ++ i)
		{
			float s = src [from + i];
			if (s >= FLOAT.MAX_UINT)
				dest [to + i] = INT.MAX_UINT;
			else if (s >= FLOAT.MAX_SINT)
				dest [to + i] = (int) (long) s;
			else if (s > 0)
				dest [to + i] = (int) s;
			else
				dest [to + i] = 0;
		}
		return dest;
	}



	public final static int[] toInts
	(
		short[] src,
		int from,
		int[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new int [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = src [from + i] & INT.MAX_USHORT;
		return dest;
	}



	public final static int[] toInts
	(
		byte[] src,
		int from,
		int[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new int [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = src [from + i] & INT.MAX_UBYTE;
		return dest;
	}



	public final static short[] toShorts
	(
		double[] src,
		int from,
		short[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new short [to + amt];
		for (int i = 0; i < amt; ++ i)
		{
			int d = (int) src [from + i];
			if (d > INT.MAX_USHORT)
				dest [to + i] = SHORT.MAX_USHORT;
			else if (d < INT.MIN_USHORT)
				dest [to + i] = SHORT.MIN_USHORT;
			else
				dest [to + i] = (short) d;
		}
		return dest;
	}



	public final static short[] toShorts
	(
		float[] src,
		int from,
		short[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new short [to + amt];
		for (int i = 0; i < amt; ++ i)
		{
			int d = (int) src [from + i];
			if (d > INT.MAX_USHORT)
				dest [to + i] = SHORT.MAX_USHORT;
			else if	(d < INT.MIN_USHORT)
				dest [to + i] = SHORT.MIN_USHORT;
			else
				dest [to + i] = (short) d;
		}
		return dest;
	}



	public final static short[] toShorts
	(
		byte[] src,
		int from,
		short[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new short [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (short) (src [from + i] & INT.MAX_UBYTE);
		return dest;
	}



	public final static byte[] toBytes
	(
		double[] src,
		int from,
		byte[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new byte [to + amt];
		for (int i = 0; i < amt; ++ i)
		{
			int d = (int) src [from + i];
			if (d > INT.MAX_UBYTE)
				dest [to + i] = BYTE.MAX_UBYTE;
			else if (d < INT.MIN_UBYTE)
				dest [to + i] = BYTE.MIN_UBYTE;
			else
				dest [to + i] = (byte) d;
		}
		return dest;
	}



	public final static byte[] toBytes
	(
		float[] src,
		int from,
		byte[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new byte [to + amt];
		for (int i = 0; i < amt; ++ i)
		{
			int d = (int) src [from + i];
			if (d > INT.MAX_UBYTE)
				dest [to + i] = BYTE.MAX_UBYTE;
			else if (d < INT.MIN_UBYTE)
				dest [to + i] = BYTE.MIN_UBYTE;
			else
				dest [to + i] = (byte) d;
		}
		return dest;
	}



	public final static char[] toChars
	(
		double[] src,
		int from,
		char[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new char [to + amt];
		for (int i = 0; i < amt; ++ i)
		{
			int d = (int) src [from + i];
			if (d > INT.MAX_CHAR)
				dest [to + i] = CHAR.MAX_CHAR;
			else if (d < INT.MIN_CHAR)
				dest [to + i] = CHAR.MIN_CHAR;
			else
				dest [to + i] = (char) d;
		}
		return dest;
	}



	public final static char[] toChars
	(
		float[] src,
		int from,
		char[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new char [to + amt];
		for (int i = 0; i < amt; ++ i)
		{
			int d = (int) src [from + i];
			if (d > INT.MAX_CHAR)
				dest [to + i] = CHAR.MAX_CHAR;
			else if (d < INT.MIN_CHAR)
				dest [to + i] = CHAR.MIN_CHAR;
			else
				dest [to + i] = (char) d;
		}
		return dest;
	}



	public final static char[] toChars
	(
		byte[] src,
		int from,
		char[] dest,
		int to,
		int amt
	)
	{
		if (amt < 0)
			throw new IllegalArgumentException ("amt==" + amt);
		if (src == null)
			return null;
		if (dest == null)
			dest = new char [to + amt];
		for (int i = 0; i < amt; ++ i)
			dest [to + i] = (char) (src [from + i] & INT.MAX_UBYTE);
		return dest;
	}



}



