///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//  Notice to licensees:                                                     //
//                                                                           //
//  This source code is the exclusive, proprietary intellectual property of  //
//  Sharkysoft (sharkysoft.com).  You may view this source code as a         //
//  supplement to other product documentation, but you may not distribute    //
//  it or use it for any other purpose without written consent from          //
//  Sharkysoft.                                                              //
//                                                                           //
//  You are permitted to modify and recompile this source code, but you may  //
//  not remove this notice.  If you add features to or fix errors in this    //
//  code, please consider sharing your changes with Sharkysoft for possible  //
//  incorporation into future releases of the product.  Thanks!              //
//                                                                           //
//  For more information about Sharkysoft products and services, please      //
//  visit Sharkysoft on the web at                                           //
//                                                                           //
//       http://sharkysoft.com/                                              //
//                                                                           //
//  Thank you for using Lava!                                                //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////



package lava.security.des;



import lava.util.BitTwiddler;
import lava.konst.INT;
import lava.konst.LONG;
import lava.security.IBlockCipher64;



/******************************************************************************
DES decryption in Cipher Block Chaining (CBC) mode.

<p><b>Details:</b>  <code>DesCbcEncoder</code> performs DES decryption in Cipher Block Chaining (CBC) mode.  See superclass documention for more details.</p>

@see DesCbcEncoder
@author Sharky
@since 1999.08.16
******************************************************************************/

public class DesCbcDecoder
	extends DesCbcCoder
{



	/**********************************************************************
	Specifies initialization vector and encryption function.

	<p><b>Details:</b>  This constructor initializes a new instance with the given block cipher and initialization vector.</p>

	@param engine the block cipher
	@param iv the initialization vector
	**********************************************************************/

	public DesCbcDecoder (IBlockCipher64 engine, long iv)
	{
		super (engine, iv);
		beginTransform ();
	}



	/**********************************************************************
	Specifies encryption key and initialization vector.

	<p><b>Details:</b>  This constructor initializes a standard DES CBC cipher with the given initialization vector and DES encryption key.</p>

	@param key the encryption key
	@param iv the initialization vector
	**********************************************************************/

	public DesCbcDecoder (DesKey key, long iv)
	{
		this (new DesEngine (key), iv);
	}


	public final synchronized void beginTransform ()
	{
		civ = iv;
	}



	private long civ;



	private final long[] desin = new long [1];



	private final long[] desout = new long [1];



	/**********************************************************************
	Decrypts one block.

	<p><b>Details:</b>  This method decrypts a single 64-bit block.</p>

	@param text the block to encrypt
	@return the encrypted result
	**********************************************************************/

	private long continueTransform (long text)
	{
		desin [0] = text;
		engine . decrypt (desin, desout, 1);
		text = desout [0] ^ civ;
		civ = desin [0];
		return text;
	}



	public final synchronized int continueTransform (Object input, Object output, int length)
	{
		if (! (input instanceof long[]))
			throw new IllegalArgumentException ("input");
		if (! (output instanceof long[]))
			throw new IllegalArgumentException ("output");
		if (length < 0)
			return 0;
		for (int i = 0; i < length; ++ i)
			((long[]) output) [i] = continueTransform (((long[]) input) [i]);
		return length;
	}



	/**********************************************************************
	Encrypts blocks using byte arrays.

	<p><b>Details:</b>  This method is not part of the <code>ITextTransform</code> API, but it is provided to facilitate symbol processing where the symbols are stored in big endian byte arrays.</p>

	<p>The length parameter and return value count <code>byte</code>s, not 64-bit blocks.  Results are undefined if <var>length</var> is positive and not a multiple of 8.</p>

	@param input the input blocks
	@param output buffer for output blocks
	@param length number of input array elements to process
	@return number of bytes written to output array
	**********************************************************************/

	public final synchronized int continueTransform (byte[] input, byte[] output, int length)
	{
		if (length < 0)
			return 0;
		int i;
		for (i = 0; i < length; i += 8)
			BitTwiddler.toBytesBe (continueTransform (BitTwiddler.toLongBe (input, i)), output, i);
		return i;
	}



}



