///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//  Notice to licensees:                                                     //
//                                                                           //
//  This source code is the exclusive, proprietary intellectual property of  //
//  Sharkysoft (sharkysoft.com).  You may view this source code as a         //
//  supplement to other product documentation, but you may not distribute    //
//  it or use it for any other purpose without written consent from          //
//  Sharkysoft.                                                              //
//                                                                           //
//  You are permitted to modify and recompile this source code, but you may  //
//  not remove this notice.  If you add features to or fix errors in this    //
//  code, please consider sharing your changes with Sharkysoft for possible  //
//  incorporation into future releases of the product.  Thanks!              //
//                                                                           //
//  For more information about Sharkysoft products and services, please      //
//  visit Sharkysoft on the web at                                           //
//                                                                           //
//       http://sharkysoft.com/                                              //
//                                                                           //
//  Thank you for using Lava!                                                //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////



package lava.security.des;



import lava.array.ArrayCaster;
import lava.konst.LONG;
import lava.security.IBlockCipher64;



/******************************************************************************
Optimized DES implementation.

<p><b>Details:</b>  DesEngine is an efficient implementation of the DES private key encryption/decryption algorithm described in FIPS PUB 46.</p>

<p>DES is a 64-bit block cipher, which means that it processes data in 64-bit chunks.  This implementation allows multiple blocks to be loaded into an array so that the blocks can be processed with a single method call.  The encrypted/<wbr>decrypted results can be returned in the same or different array.</p>

<p><b>initial and final permutations</b></p>

<p>Optionally, if one is willing to sacrifice perfect conformance to the DES standard in exchange for faster processing, individual instances of <code>DesEngine</code> can be configured to skip the initial and final permutations.  These permutations do not contribute to the security of DES but <em>are</em> part of the standard.</p>

<p><b>bit ordering</b></p>

<p><strong>Refer to the package summary for important information regarding bit ordering.</strong></p>

<p><b>usage</b></p>

<p><code>DesEngine</code> is an implementation of <code>IBlockCipher64</code>.  <strong>You must understand the contract specified by that interface before you can have any success using this class.  Please read it!</strong></p>

@since 1999.08.03
@version 2001.03.09
******************************************************************************/

public final class DesEngine
	implements IBlockCipher64
{



	/**********************************************************************
	Key.
	**********************************************************************/

	private final DesKey key;



	/**********************************************************************
	Initial text permutations enabled.

	<p><b>Details:</b>  This variable records whether or not an initial text permutation should be performed prior to encoding/decoding a block.</p>
	**********************************************************************/

	private final boolean itpe;



	/**********************************************************************
	Final text permutations enabled.

	<p><b>Details:</b>  This variable records whether or not a final text permutation should be performed prior to encoding/decoding a block.</p>
	**********************************************************************/

	private final boolean ftpe;



	/**********************************************************************
	Sets key; enables/disables permutations.

	<p><b>Details:</b>  This constructor initializes a new <code>DesEngine</code> instance and enables or disables the initial and final text permutations.</p>

	@param key the encryption key
	@param itpe initial text permutation enabled
	@param ftpe final text permutation enabled
	**********************************************************************/

	public DesEngine (DesKey key, boolean itpe, boolean ftpe)
	{
		this . key = key;
		this . itpe = itpe;
		this . ftpe = ftpe;
	}



	/**********************************************************************
	Sets key.

	<p><b>Details:</b>  This constructor initializes a new <code>DesEngine</code> instance to perform standard DES encryption/decryption using the specified key.</p>

	@param key the encryption key
	@since 1999.08.17
	**********************************************************************/

	public DesEngine (DesKey key)
	{
		this (key, true, true);
	}



	/**********************************************************************
	Mix operation.

	<p><b>Details:</b>  This method performs the mix operation on <var>mixr</var>, the right half of a DES round, using the given subkey (<var>skl</var> and <var>skr</var>).</p>

	@param mixr the mix input
	@param skl left half of subkey
	@param skr right half of subkey
	@return the mixed result
	**********************************************************************/

	private static int mix (int mixr, int skl, int skr)
	{
		// text expansion permutation
		int mixl =
			(mixr & 0x001f8000) >>> 15
		|	(mixr & 0x01f80000) >>> 13
		|	(mixr & 0x1f800000) >>> 11
		|	(mixr & 0xf8000000) >>>  9
		|	(mixr & 0x00000001)  << 23;
		mixr =
			(mixr & 0x80000000) >>> 31
		|	(mixr & 0x0000001f)  <<  1
		|	(mixr & 0x000001f8)  <<  3
		|	(mixr & 0x00001f80)  <<  5
		|	(mixr & 0x0001f800)  <<  7;
		// xor with subkey
		mixl ^= skl;
		mixr ^= skr;
		// s-box application and p-box permutation
		mixr =
			DesSPBoxes.values [0] [ mixr         & 0x3f]
		|	DesSPBoxes.values [1] [(mixr >>>  6) & 0x3f]
		|	DesSPBoxes.values [2] [(mixr >>> 12) & 0x3f]
		|	DesSPBoxes.values [3] [(mixr >>> 18) & 0x3f]
		|	DesSPBoxes.values [4] [ mixl         & 0x3f]
		|	DesSPBoxes.values [5] [(mixl >>>  6) & 0x3f]
		|	DesSPBoxes.values [6] [(mixl >>> 12) & 0x3f]
		|	DesSPBoxes.values [7] [(mixl >>> 18) & 0x3f];
		return mixr;
	}



	/**********************************************************************
	Indicates forward DES function (encryption).
	**********************************************************************/

	private static final boolean FORWARD = true;



	/**********************************************************************
	Indicates inverse DES function (decryption).
	**********************************************************************/

	private static final boolean REVERSE = false;



	/**********************************************************************
	Encrypts blocks.

	<p><b>Details:</b></p>  See IBlockCipher64 (read all of it!) for a description of this interface method's contract.</p>

	@param src the source array
	@param dest the destination array
	@param length the number of elements in the source array
	**********************************************************************/

	public final void encrypt (Object src, Object dest, int length)
	{
		crypt (FORWARD, src, dest, length);
	}



	/**********************************************************************
	Decrypts blocks.

	<p><b>Details:</b></p>  See IBlockCipher64 (read all of it!) for a description of this interface method's contract.</p>

	@param src the source array
	@param dest the destination array
	@param length the number of elements in the source array
	**********************************************************************/

	public final void decrypt (Object src, Object dest, int length)
	{
		crypt (REVERSE, src, dest, length);
	}



	/**********************************************************************
	Encrypts or decrypts text.

	<p><b>Details:</b> This method encrypts or deccrypts the given text array, depending on the value of <var>forward</var>.</p>

	@param forward whether to encrypt or decrypt
	@param input the source array
	@param output the destination array
	@param length length of array
	**********************************************************************/

	// not "static" because itpe and ftpe are referenced
	private void crypt
	(
		boolean forward,
		Object input,
		Object output,
		int length
	)
	{
		if (length < 0)
			throw new IllegalArgumentException ("length=" + length);
		// Factored out for speed!
		if (input instanceof byte[])
			if (output instanceof byte[])
				crypt (forward, (byte []) input, (byte []) output, length);
			else if (output instanceof int[])
				crypt (forward, (byte []) input, (int  []) output, length);
			else if (output instanceof long[])
				crypt (forward, (byte []) input, (long []) output, length);
			else if (output instanceof short[])
				crypt (forward, (byte []) input, (short[]) output, length);
			else
				throw new IllegalArgumentException ("output");
		else if (input instanceof int[])
			if (output instanceof int[])
				crypt (forward, (int  []) input, (int  []) output, length);
			else if (output instanceof byte[])
				crypt (forward, (int  []) input, (byte []) output, length);
			else if (output instanceof long[])
				crypt (forward, (int  []) input, (long []) output, length);
			else if (output instanceof short[])
				crypt (forward, (int  []) input, (short[]) output, length);
			else
				throw new IllegalArgumentException ("output");
		else if (input instanceof long[])
			if (output instanceof long[])
				crypt (forward, (long []) input, (long []) output, length);
			else if (output instanceof int[])
				crypt (forward, (long []) input, (int  []) output, length);
			else if (output instanceof byte[])
				crypt (forward, (long []) input, (byte []) output, length);
			else if (output instanceof short[])
				crypt (forward, (long []) input, (short[]) output, length);
			else
				throw new IllegalArgumentException ("output");
		else if (input instanceof short[])
			if (output instanceof short[])
				crypt (forward, (short[]) input, (short[]) output, length);
			else if (output instanceof byte[])
				crypt (forward, (short[]) input, (byte []) output, length);
			else if (output instanceof int[])
				crypt (forward, (short[]) input, (int  []) output, length);
			else if (output instanceof long[])
				crypt (forward, (short[]) input, (long []) output, length);
			else
				throw new IllegalArgumentException ("output");
		else
			throw new IllegalArgumentException ("input");
	}



	private void crypt (boolean forward, long[] in64, long[] out64, int length)
	{
		for (int i = 0; i < length; ++ i)
			out64 [i] = crypt (forward, (int) (in64 [i] >> 32), (int) in64 [i]);
	}



	private void crypt (boolean forward, long[] in64, int[] out32, int length)
	{
		long text;
		int o = 0;
		for (int i = 0; i < length; i += 1, o += 2)
		{
			text = crypt (forward, (int) (in64 [i] >> 32), (int) in64 [i]);
			out32 [o    ] = (int) (text >> 32);
			out32 [o + 1] = (int)  text       ;
		}
	}



	private void crypt (boolean forward, long[] in64, short[] out16, int length)
	{
		int l0, r0;
		long text;
		int o = 0;
		for (int i = 0; i < length; i += 1, o += 4)
		{
			text = crypt (forward, (int) (in64 [i] >> 32), (int) in64 [i]);
			l0 = (int) (text >> 32);
			r0 = (int)  text       ;
			out16 [o    ] = (short) (l0 >> 16);
			out16 [o + 1] = (short)  l0       ;
			out16 [o + 2] = (short) (r0 >> 16);
			out16 [o + 3] = (short)  r0       ;
		}
	}



	private void crypt (boolean forward, long[] in64, byte[] out08, int length)
	{
		int l0, r0;
		long text;
		int o = 0;
		for (int i = 0; i < length; i += 1, o += 8)
		{
			text = crypt (forward, (int) (in64 [i] >> 32), (int) in64 [i]);
			l0 = (int) (text >> 32);
			r0 = (int)  text       ;
			out08 [o    ] = (byte) (l0 >> 24);
			out08 [o + 1] = (byte) (l0 >> 16);
			out08 [o + 2] = (byte) (l0 >>  8);
			out08 [o + 3] = (byte)  l0       ;
			out08 [o + 4] = (byte) (r0 >> 24);
			out08 [o + 5] = (byte) (r0 >> 16);
			out08 [o + 6] = (byte) (r0 >>  8);
			out08 [o + 7] = (byte)  r0       ;
		}
	}



	private void crypt (boolean forward, int[] in32, long[] out64, int length)
	{
		int o = 0;
		for (int i = 0; i < length; i += 2, o += 1)
			out64 [o] = crypt (forward, in32 [i], in32 [i + 1]);
	}



	private void crypt (boolean forward, int[] in32, int[] out32, int length)
	{
		long text;
		for (int i = 0; i < length; i += 2)
		{
			text = crypt (forward, in32 [i], in32 [i + 1]);
			out32 [i    ] = (int) (text >> 32);
			out32 [i + 1] = (int)  text       ;
		}
	}



	private void crypt (boolean forward, int[] in32, short[] out16, int length)
	{
		int l0, r0;
		long text;
		int o = 0;
		for (int i = 0; i < length; i += 2, o += 4)
		{
			text = crypt (forward, in32 [i], in32 [i + 1]);
			l0 = (int) (text >> 32);
			r0 = (int)  text       ;
			out16 [o    ] = (short) (l0 >> 16);
			out16 [o + 1] = (short)  l0       ;
			out16 [o + 2] = (short) (r0 >> 16);
			out16 [o + 3] = (short)  r0       ;
		}
	}



	private void crypt (boolean forward, int[] in32, byte[] out08, int length)
	{
		int l0, r0;
		long text;
		int o = 0;
		for (int i = 0; i < length; i += 2, o += 8)
		{
			text = crypt (forward, in32 [i], in32 [i + 1]);
			l0 = (int) (text >> 32);
			r0 = (int)  text       ;
			out08 [o    ] = (byte) (l0 >> 24);
			out08 [o + 1] = (byte) (l0 >> 16);
			out08 [o + 2] = (byte) (l0 >>  8);
			out08 [o + 3] = (byte)  l0       ;
			out08 [o + 4] = (byte) (r0 >> 24);
			out08 [o + 5] = (byte) (r0 >> 16);
			out08 [o + 6] = (byte) (r0 >>  8);
			out08 [o + 7] = (byte)  r0       ;
		}
	}




	private void crypt (boolean forward, short[] in16, long[] out64, int length)
	{
		int l0, r0;
		long text;
		int o = 0;
		for (int i = 0; i < length; i += 4, o += 1)
			out64 [o] = crypt
			(	forward,
				in16 [i    ] << 16 | (in16 [i + 1] & 0xffff),
				in16 [i + 2] << 16 | (in16 [i + 3] & 0xffff)
			);
	}



	private void crypt (boolean forward, short[] in16, int[] out32, int length)
	{
		int l0, r0;
		long text;
		int o = 0;
		for (int i = 0; i < length; i += 4, o += 2)
		{
			text = crypt
			(	forward,
				in16 [i    ] << 16 | (in16 [i + 1] & 0xffff),
				in16 [i + 2] << 16 | (in16 [i + 3] & 0xffff)
			);
			out32 [o    ] = (int) (text >> 32);
			out32 [o + 1] = (int)  text       ;
		}
	}



	private void crypt (boolean forward, short[] in16, short[] out16, int length)
	{
		int l0, r0;
		long text;
		int o = 0;
		for (int i = 0; i < length; i += 4, o += 4)
		{
			text = crypt
			(	forward,
				in16 [i    ] << 16 | (in16 [i + 1] & 0xffff),
				in16 [i + 2] << 16 | (in16 [i + 3] & 0xffff)
			);
			l0 = (int) (text >> 32);
			r0 = (int)  text       ;
			out16 [o    ] = (short) (l0 >> 16);
			out16 [o + 1] = (short)  l0       ;
			out16 [o + 2] = (short) (r0 >> 16);
			out16 [o + 3] = (short)  r0       ;
		}
	}



	private void crypt (boolean forward, short[] in16, byte[] out08, int length)
	{
		int l0, r0;
		long text;
		int o = 0;
		for (int i = 0; i < length; i += 4, o += 8)
		{
			text = crypt
			(	forward,
				in16 [i    ] << 16 | (in16 [i + 1] & 0xffff),
				in16 [i + 2] << 16 | (in16 [i + 3] & 0xffff)
			);
			l0 = (int) (text >> 32);
			r0 = (int)  text       ;
			out08 [o    ] = (byte) (l0 >> 24);
			out08 [o + 1] = (byte) (l0 >> 16);
			out08 [o + 2] = (byte) (l0 >>  8);
			out08 [o + 3] = (byte)  l0       ;
			out08 [o + 4] = (byte) (r0 >> 24);
			out08 [o + 5] = (byte) (r0 >> 16);
			out08 [o + 6] = (byte) (r0 >>  8);
			out08 [o + 7] = (byte)  r0       ;
		}
	}



	private void crypt (boolean forward, byte[] in08, long[] out64, int length)
	{
		int o = 0;
		for (int i = 0; i < length; i += 8, o += 1)
			out64 [o] = crypt
			(	forward,
				 in08 [i    ]         << 24
			|	(in08 [i + 1] & 0xff) << 16
			|	(in08 [i + 2] & 0xff) <<  8
			|	(in08 [i + 3] & 0xff)      ,
				 in08 [i + 4]         << 24
			|	(in08 [i + 5] & 0xff) << 16
			|	(in08 [i + 6] & 0xff) <<  8
			|	(in08 [i + 7] & 0xff)
			);
	}



	private void crypt (boolean forward, byte[] in08, int[] out32, int length)
	{
		long text;
		int o = 0;
		for (int i = 0; i < length; i += 8, o += 2)
		{
			text = crypt
			(	forward,
				 in08 [i    ]         << 24
			|	(in08 [i + 1] & 0xff) << 16
			|	(in08 [i + 2] & 0xff) <<  8
			|	(in08 [i + 3] & 0xff)      ,
				 in08 [i + 4]         << 24
			|	(in08 [i + 5] & 0xff) << 16
			|	(in08 [i + 6] & 0xff) <<  8
			|	(in08 [i + 7] & 0xff)
			);
			out32 [o    ] = (int) (text >> 32);
			out32 [o + 1] = (int)  text       ;
		}
	}



	private void crypt (boolean forward, byte[] in08, short[] out16, int length)
	{
		int l0, r0;
		long text;
		int o = 0;
		for (int i = 0; i < length; i += 8, o += 4)
		{
			text = crypt
			(	forward,
				 in08 [i    ]         << 24
			|	(in08 [i + 1] & 0xff) << 16
			|	(in08 [i + 2] & 0xff) <<  8
			|	(in08 [i + 3] & 0xff)      ,
				 in08 [i + 4]         << 24
			|	(in08 [i + 5] & 0xff) << 16
			|	(in08 [i + 6] & 0xff) <<  8
			|	(in08 [i + 7] & 0xff)
			);
			l0 = (int) (text >> 32);
			r0 = (int)  text       ;
			out16 [o    ] = (short) (l0 >> 16);
			out16 [o + 1] = (short)  l0       ;
			out16 [o + 2] = (short) (r0 >> 16);
			out16 [o + 3] = (short)  r0       ;
		}
	}



	private void crypt (boolean forward, byte[] in08, byte[] out08, int length)
	{
		int l0, r0;
		long text;
		int o = 0;
		for (int i = 0; i < length; i += 8, o += 8)
		{
			text = crypt
			(	forward,
				 in08 [i    ]         << 24
			|	(in08 [i + 1] & 0xff) << 16
			|	(in08 [i + 2] & 0xff) <<  8
			|	(in08 [i + 3] & 0xff)      ,
				 in08 [i + 4]         << 24
			|	(in08 [i + 5] & 0xff) << 16
			|	(in08 [i + 6] & 0xff) <<  8
			|	(in08 [i + 7] & 0xff)
			);
			l0 = (int) (text >> 32);
			r0 = (int)  text       ;
			out08 [o    ] = (byte) (l0 >> 24);
			out08 [o + 1] = (byte) (l0 >> 16);
			out08 [o + 2] = (byte) (l0 >>  8);
			out08 [o + 3] = (byte)  l0       ;
			out08 [o + 4] = (byte) (r0 >> 24);
			out08 [o + 5] = (byte) (r0 >> 16);
			out08 [o + 6] = (byte) (r0 >>  8);
			out08 [o + 7] = (byte)  r0       ;
		}
	}




	/**********************************************************************
	Encrypts or decrypts symbol.

	<p><b>Details:</b>  This method encrypts or decrypts the given input symbol and returns the output symbol.</p>

	@param forward whether to encrypt or decrypt
	@param l0 left half of input symbol
	@param r0 right half of input symbol
	@return the output symbol

	@since 1999.08.28
	@version 2001.03.09
	**********************************************************************/

	long crypt (boolean forward, int l0, int r0)
	{
		int l1, r1;
		// Perform initial text permutation:
		if (forward ? itpe : ftpe)
		{
			l1 =
				(l0 & 0x01000000) >>> 24
			|	(r0 & 0x01000000) >>> 20
			|	(l0 & 0x04000000) >>> 18
			|	(l0 & 0x00010000) >>> 15
			|	(r0 & 0x04000000) >>> 14
			|	(l0 & 0x10000000) >>> 12
			|	(r0 & 0x00010000) >>> 11
			|	(l0 & 0x00040000) >>>  9
			|	(r0 & 0x10000000) >>>  8
			|	(l0 & 0x40000100) >>>  6
			|	(r0 & 0x00040000) >>>  5
			|	(l0 & 0x00100000) >>>  3
			|	(r0 & 0x40000100) >>>  2
			|	 l0 & 0x00000400
			|	(r0 & 0x00100000)  <<  1
			|	(l0 & 0x00400001)  <<  3
			|	(r0 & 0x00000400)  <<  4
			|	(l0 & 0x00001000)  <<  6
			|	(r0 & 0x00400001)  <<  7
			|	(l0 & 0x00000004)  <<  9
			|	(r0 & 0x00001000)  << 10
			|	(l0 & 0x00004000)  << 12
			|	(r0 & 0x00000004)  << 13
			|	(l0 & 0x00000010)  << 15
			|	(r0 & 0x00004000)  << 16
			|	(r0 & 0x00000010)  << 19
			|	(l0 & 0x00000040)  << 21
			|	(r0 & 0x00000040)  << 25;
			r1 =
				(l0 & 0x02000000) >>> 25
			|	(r0 & 0x02000000) >>> 21
			|	(l0 & 0x08000000) >>> 19
			|	(l0 & 0x00020000) >>> 16
			|	(r0 & 0x08000000) >>> 15
			|	(l0 & 0x20000000) >>> 13
			|	(r0 & 0x00020000) >>> 12
			|	(l0 & 0x00080000) >>> 10
			|	(r0 & 0x20000000) >>>  9
			|	(l0 & 0x80000200) >>>  7
			|	(r0 & 0x00080000) >>>  6
			|	(l0 & 0x00200000) >>>  4
			|	(r0 & 0x80000200) >>>  3
			|	(l0 & 0x00000800) >>>  1
			|	 r0 & 0x00200000
			|	(l0 & 0x00800002)  <<  2
			|	(r0 & 0x00000800)  <<  3
			|	(l0 & 0x00002000)  <<  5
			|	(r0 & 0x00800002)  <<  6
			|	(l0 & 0x00000008)  <<  8
			|	(r0 & 0x00002000)  <<  9
			|	(l0 & 0x00008000)  << 11
			|	(r0 & 0x00000008)  << 12
			|	(l0 & 0x00000020)  << 14
			|	(r0 & 0x00008000)  << 15
			|	(r0 & 0x00000020)  << 18
			|	(l0 & 0x00000080)  << 20
			|	(r0 & 0x00000080)  << 24;
			l0 = l1;
			r0 = r1;
		}
		// Do rounds:
		int[] ckl, ckr;
		if (forward)
		{
			ckl = key . ekl;
			ckr = key . ekr;
		}
		else
		{
			ckl = key . dkl;
			ckr = key . dkr;
		}
		l1 = r0;
		r1 = l0 ^ mix (r0, ckl [0], ckr [0]);
		l0 = r1;
		r0 = l1 ^ mix (r1, ckl [1], ckr [1]);
		l1 = r0;
		r1 = l0 ^ mix (r0, ckl [2], ckr [2]);
		l0 = r1;
		r0 = l1 ^ mix (r1, ckl [3], ckr [3]);
		l1 = r0;
		r1 = l0 ^ mix (r0, ckl [4], ckr [4]);
		l0 = r1;
		r0 = l1 ^ mix (r1, ckl [5], ckr [5]);
		l1 = r0;
		r1 = l0 ^ mix (r0, ckl [6], ckr [6]);
		l0 = r1;
		r0 = l1 ^ mix (r1, ckl [7], ckr [7]);
		l1 = r0;
		r1 = l0 ^ mix (r0, ckl [8], ckr [8]);
		l0 = r1;
		r0 = l1 ^ mix (r1, ckl [9], ckr [9]);
		l1 = r0;
		r1 = l0 ^ mix (r0, ckl [10], ckr [10]);
		l0 = r1;
		r0 = l1 ^ mix (r1, ckl [11], ckr [11]);
		l1 = r0;
		r1 = l0 ^ mix (r0, ckl [12], ckr [12]);
		l0 = r1;
		r0 = l1 ^ mix (r1, ckl [13], ckr [13]);
		l1 = r0;
		r1 = l0 ^ mix (r0, ckl [14], ckr [14]);
		l0 = r1;
		r0 = l1 ^ mix (r1, ckl [15], ckr [15]);
		// Perform final_text_permutation:
		if (forward ? ftpe : itpe)
		{
			l1 =
				(r0 & 0x08000000) >>> 21
			|	(l0 & 0x08000000) >>> 20
			|	(r0 & 0x00080000) >>> 15
			|	(l0 & 0x00080000) >>> 14
			|	(r0 & 0x04000000) >>> 12
			|	(l0 & 0x04000000) >>> 11
			|	(r0 & 0x00000800) >>>  9
			|	(l0 & 0x00000800) >>>  8
			|	(r0 & 0x00040000) >>>  6
			|	(l0 & 0x00040000) >>>  5
			|	(r0 & 0x02000008) >>>  3
			|	(l0 & 0x02000008) >>>  2
			|	 r0 & 0x00000400
			|	(l0 & 0x00000400)  <<  1
			|	(r0 & 0x00020000)  <<  3
			|	(l0 & 0x00020000)  <<  4
			|	(r0 & 0x01000004)  <<  6
			|	(l0 & 0x01000004)  <<  7
			|	(r0 & 0x00000200)  <<  9
			|	(l0 & 0x00000200)  << 10
			|	(r0 & 0x00010000)  << 12
			|	(l0 & 0x00010000)  << 13
			|	(r0 & 0x00000002)  << 15
			|	(l0 & 0x00000002)  << 16
			|	(r0 & 0x00000100)  << 18
			|	(l0 & 0x00000100)  << 19
			|	(r0 & 0x00000001)  << 24
			|	(l0 & 0x00000001)  << 25;
			r1 =
				(r0 & 0x80000000) >>> 25
			|	(l0 & 0x80000000) >>> 24
			|	(r0 & 0x00800000) >>> 19
			|	(l0 & 0x00800000) >>> 18
			|	(r0 & 0x40000000) >>> 16
			|	(l0 & 0x40000000) >>> 15
			|	(r0 & 0x00008000) >>> 13
			|	(l0 & 0x00008000) >>> 12
			|	(r0 & 0x00400000) >>> 10
			|	(l0 & 0x00400000) >>>  9
			|	(r0 & 0x20000080) >>>  7
			|	(l0 & 0x20000080) >>>  6
			|	(r0 & 0x00004000) >>>  4
			|	(l0 & 0x00004000) >>>  3
			|	(r0 & 0x00200000) >>>  1
			|	 l0 & 0x00200000
			|	(r0 & 0x10000040)  <<  2
			|	(l0 & 0x10000040)  <<  3
			|	(r0 & 0x00002000)  <<  5
			|	(l0 & 0x00002000)  <<  6
			|	(r0 & 0x00100000)  <<  8
			|	(l0 & 0x00100000)  <<  9
			|	(r0 & 0x00000020)  << 11
			|	(l0 & 0x00000020)  << 12
			|	(r0 & 0x00001000)  << 14
			|	(l0 & 0x00001000)  << 15
			|	(r0 & 0x00000010)  << 20
			|	(l0 & 0x00000010)  << 21;
		}
		else
		{
			l1 = r0;
			r1 = l0;
		}
		return (long) l1 << 32 | r1 & LONG.MAX_UINT;
	}



}



