///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//  Notice to licensees:                                                     //
//                                                                           //
//  This source code is the exclusive, proprietary intellectual property of  //
//  Sharkysoft (sharkysoft.com).  You may view this source code as a         //
//  supplement to other product documentation, but you may not distribute    //
//  it or use it for any other purpose without written consent from          //
//  Sharkysoft.                                                              //
//                                                                           //
//  You are permitted to modify and recompile this source code, but you may  //
//  not remove this notice.  If you add features to or fix errors in this    //
//  code, please consider sharing your changes with Sharkysoft for possible  //
//  incorporation into future releases of the product.  Thanks!              //
//                                                                           //
//  For more information about Sharkysoft products and services, please      //
//  visit Sharkysoft on the web at                                           //
//                                                                           //
//       http://sharkysoft.com/                                              //
//                                                                           //
//  Thank you for using Lava!                                                //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////



package lava.security.des;



import lava.array.ArrayCaster;
import lava.konst.LONG;
import lava.security.IBlockCipher64;



/******************************************************************************
Encrypts data using DES in cipher feedback (CFB) mode.

<p><b>Details:</b>  <code>DesOfbCoder</code> performs data encryption/decryption using DES in output feedback (CFB) mode.  This class performs both encryption and decryption, because the two operations are identical and self-<wbr>canceling.  This special mode of DES is described in FIPS PUB 81.  Refer to the superclass documentation and for details.</p>

@since 1999.08.23
@author Sharky
******************************************************************************/

public class DesOfbCoder
	extends DesFbCoder
{



	/**********************************************************************
	Initializes with arbitrary block cipher, value for k, and initialization vector.

	<p><b>Details:</b>  This constructor initializes a new instance to encrypt/<wbr>decrypt k-bit symbols using the given 64-bit block cipher and initialization vector (IV).</p>

	@param engine the block cipher
	@param k the symbol width
	@param iv the initialization vector
	**********************************************************************/

	public DesOfbCoder (IBlockCipher64 engine, int k, long iv)
	{
		super (engine, k, iv);
	}



	/**********************************************************************
	Initializes with DES function, value for k, and initialization vector.

	<p><b>Details:</b>  This constructor initializes a new instance to encrypt/<wbr>decrypt k-bit symbols using the DES encryption function corresponding to the given key and initialization vector (IV).</p>

	@param engine the block cipher
	@param k the symbol width
	@param iv the initialization vector
	**********************************************************************/

	public DesOfbCoder (DesKey key, int k, long iv)
	{
		this (new DesEngine (key), k, iv);
	}



	/**********************************************************************
	ITextTransform interface method.

	<p><b>Details:</b>  Refer to <code>ITextTransform</code> for a description of this interface method's contract.</p>

	<p>This implementation performs both encryption and decryption, immediately producing one output symbol for each input symbol.</p>

	@param input the input symbols
	@param output buffer for output symbols
	@param len number of input symbols to process
	@return number of output symbols written to <var>output</var>
	**********************************************************************/

	public synchronized int continueTransform (Object input, Object output, int len)
	{
		final int inwidth = ArrayCaster.getPrimitiveElementType (input) * 8;
		if (inwidth < k)
			throw new IllegalArgumentException ("input");
		final int outwidth = ArrayCaster.getPrimitiveElementType (output) * 8;
		if (outwidth < k)
			throw new IllegalArgumentException ("output");
		long text;
		for (int i = 0; i < len; ++ i)
		{
			engine . encrypt (desin, desout, 1);
			switch (inwidth)
			{
			case 8:
				text = ((byte[]) input) [i];
				break;
			case 16:
				text = ((short[]) input) [i];
				break;
			case 32:
				text = ((int[]) input) [i];
				break;
			case 64:
				text = ((long[]) input) [i];
				break;
			default:
				throw new IllegalArgumentException ("input");
			}
			text &= kmask;
			text = (desout [0] ^ (text << nk)) >>> nk;
			switch (outwidth)
			{
			case 8:
				((byte[]) output) [i] = (byte) text;
				break;
			case 16:
				((short[]) output) [i] = (short) text;
				break;
			case 32:
				((int[]) output) [i] = (int) text;
				break;
			case 64:
				((long[]) output) [i] = text;
				break;
			default:
				throw new IllegalArgumentException ("output");
			}
			if (k == 64)
				desin [0] = desout [0];
			else
				desin [0] = desin [0] << k | desout [0] >>> nk;
		}
		if (len < 0)
			return 0;
		return len;
	}



}



