///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//  Notice to licensees:                                                     //
//                                                                           //
//  This source code is the exclusive, proprietary intellectual property of  //
//  Sharkysoft (sharkysoft.com).  You may view this source code as a         //
//  supplement to other product documentation, but you may not distribute    //
//  it or use it for any other purpose without written consent from          //
//  Sharkysoft.                                                              //
//                                                                           //
//  You are permitted to modify and recompile this source code, but you may  //
//  not remove this notice.  If you add features to or fix errors in this    //
//  code, please consider sharing your changes with Sharkysoft for possible  //
//  incorporation into future releases of the product.  Thanks!              //
//                                                                           //
//  For more information about Sharkysoft products and services, please      //
//  visit Sharkysoft on the web at                                           //
//                                                                           //
//       http://sharkysoft.com/                                              //
//                                                                           //
//  Thank you for using Lava!                                                //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////



package lava.security.des.spec;



/******************************************************************************
DES permutation tables.

<p><b>Details:</b>  This class provides convenient access to all of the permutation functions described in the official DES specification.  This class defines one variable for each permutation.</p>

@since 1999.08.04
@author Sharky
******************************************************************************/

public final class DesPermutation
{



	private final int[] fipsvalues;



	private DesPermutation (int[] fipsvalues)
	{
		this . fipsvalues = fipsvalues;
	}



	/**********************************************************************
	Returns table size.

	<p><b>Details:</b>  This method returns the number of entries in the permutation table.</p>

	@return the size
	**********************************************************************/

	public final int getTableSize ()
	{
		return fipsvalues . length;
	}



	/**********************************************************************
	Returns value from permutation table.

	<p><b>Details:</b>  This method returns an entry from the table.  Both the index into the table and the returned value correspond to the FIPS bit numbering scheme.  (The leftmost bit is bit number 1.)  <var>fipsindex</var> is the index of the output bit, and the return value is the index of the input bit.</p>

	<p>For example, if <code>getFipsValue(7)</code> returns 3, then bit 7 of the permutation function's output is the same as bit 3 of the permutation function's input.</p>

	@param fipsindex the index of the output bit
	@return the index of the input bit
	**********************************************************************/

	public final int getFipsValue (int fipsindex)
	{
		return fipsvalues [fipsindex - 1];
	}



	/**********************************************************************
	Returns value from reordered permutation table.

	<p><b>Details:</b>  This method returns an entry from the "reordered" table.  By "reordered", we mean that bits are numbered as naturally as possible for software, with the right most bit indexed as "0" and the indeces increasing to the right.  <var>bitindex</var> is the index of the output bit, and the return value is the index of the input bit.</p>

	<p>For example, if <code>getFipsValue(7)</code> returns 3, then bit 7 of the permutation function's output is the same as bit 3 of the permutation function's input.</p>

	@param bitindex the index of the output bit
	@return the index of the input bit
	**********************************************************************/

	public final int getReorderedValue (int bitindex)
	{
		return getTableSize () - getFipsValue (getTableSize () - bitindex);
	}



	/**********************************************************************
	Initial key compression permutation.

	<p><b>Details:</b>  This instance describes the initial key compression permutation, identified in FIPS PUB 46-2 as "permuted choice 1" (PC-1).</p>
	**********************************************************************/

	public static final DesPermutation ikcp = new DesPermutation (DesInitialKeyCompressionPermutation.fipsvalues);



	/**********************************************************************
	Final key compression permutation.

	<p><b>Details:</b>  This instance describes the final key compression permutation, identified in FIPS PUB 46-2 as "permuted choice 2" (PC-2).</p>
	**********************************************************************/

	public static final DesPermutation fkcp = new DesPermutation (DesFinalKeyCompressionPermutation.fipsvalues);



	/**********************************************************************
	Initial text permutation.

	<p><b>Details:</b>  This instance describes the initial text permutation, which is identified in FIPS PUB 46-2 as the "initial permutation" (IP).</p>
	**********************************************************************/

	public static final DesPermutation itp = new DesPermutation (DesInitialTextPermutation.fipsvalues);



	/**********************************************************************
	Final text permutation.

	<p><b>Details:</b>  This instance describes the final text permutation, identified in FIPS PUB 46-2 as the "inverse of the initial permutation" (IP<sup>-1</sup>).</p>
	**********************************************************************/

	public static final DesPermutation ftp = new DesPermutation (DesFinalTextPermutation.fipsvalues);



	/**********************************************************************
	Text expansion permutation.

	<p><b>Details:</b>  This instance describes the text expansion permutation, identified in FIPS PUB 46-2 as the function "E" ("<i>E</i> bit-selection table").</p>
	**********************************************************************/

	public static final DesPermutation tep = new DesPermutation (DesTextExpansionPermutation.fipsvalues);




	/**********************************************************************
	P box permutation.

	<p><b>Details:</b>  This instance describes the P box permutation, identified in FIPS PUB 46-2 as the "primitive function P".</p>
	**********************************************************************/

	public static final DesPermutation pbp = new DesPermutation (DesPBoxPermutation.fipsvalues);



}



